#############################################################################
##
## Suppose that q is odd and q | p^m-1 and m is minimal with this. 
## GLNumbers returns the number of q-classes in GL(n,p)
##
GLNumbers := function(n,m)
    if m = 1 and n = 1 then 
        return 1;
    elif m = 1 and n = 2 then 
        return (q+3)/2; 
    elif m = 1 and n = 3 then 
        return (q^2 + 4*q+7 + 2 * q3)/6; 
    elif m = 1 and n = 4 then 
        return (q^3+7*q^2+21*q+19+8*q3+6*q4)/24; 
    elif m = 1 and n = 5 then 
        return (q^4+11*q^3+46*q^2+111*q+71+40*q3+30*q4+24*q5)/120;
    elif m = 2 and n = 4 then 
        return (q+3+q4)/4;
    elif m = 2 and n = 5 then 
        return (q+3+q4)/4;
    elif m = 3 and n = 5 then 
        return 1;
    elif m = n-1 and n > 2 then 
        return 1;
    elif m = n then 
        return 1;
    fi;
end;

#############################################################################
##
## b = [b1..bl]. GLZuppos returns the number of q-classes in the direct
## product of subgroups GL(bi,p)
## See Theorem YY
##
GLZuppos := function( b )
    local D, s;

    # b = [m]: group is GL(m,p)
    if Length(b) = 1 then 
        D := [D1, D2, D3, D4, D5];
        s := Sum(List([1..b[1]], x -> GLNumbers(b[1],x)*D[x]));
        return s*(1-Q2) + b[1]*Q2;

    # b = [1,1]: group is C_{p-1}^2
    elif b = [1,1] then 
        return D1*(q+1);

    # b = [1,2]: group is C_{p-1} x GL(2,p)
    elif b = [1,2] then 
        return D2 + D1*(1-Q2)*(q^2+2*q+3)/2 + Q2*5; 

    # b = [1,3]: group is C_{p-1} x GL(3,p)
    elif b = [1,3] then
        return D3 + D2 + D1*(1-Q2)*(q^3+4*q^2+6*q+7+2*q3)/6 + Q2*7;

    # b = [2,2]: group is GL(2,p)^2
    elif b = [2,2] then
        return D2*(q+3)/2 + D1*(1-Q2)*(q^3+3*q^2+5*q+7)/4 + Q2*8;

    # b = [1,1,1]: group is C_{p-1}^3
    elif b = [1,1,1] then 
        return D1*(q^2+q+1); 

#    # b = [1,1,2]: group is C_{p-1}^2 x GL(2,p)
#    elif b = [1,1,2] then 
#        return D2 + D1*(1-Q2)*(q^3+2*q^2+2*q+3)/2 + Q2*11;
    fi;
end;

#############################################################################
##
## Check if the automorphism M has a block form.
##
HasBlockForm := function( M, A )
    local d, b, i, c, s, t, B;

    # set up
    d := Length(M);
    M := M*p^0;

    # get depth of entries
    b := List(M, x -> 0);
    for i in [1..d] do
        while M[i][b[i]+1] = 0*M[i][b[i]+1] do b[i] := b[i]+1; od;
    od;

    # check 
    for i in [1..d-1] do
        if b[i]>b[i+1] then return false; fi;
    od;

    # get block sizes
    c := List(Collected(b), x -> x[2]);

    # check diagonal blocks
    s := 1;
    for i in [1..Length(c)] do
        t := s+c[i]-1;
        B := M{[s..t]}{[s..t]};
        if ForAll(Flat(B), IsConstantRationalFunction) then 
            c[i] := false;
        elif B <> A{[s..t]}{[s..t]} then 
            return false;
        fi;
        s := t+1;
    od;
  
    return SortedList(Filtered(c, x -> not IsBool(x)));
end;

#############################################################################
##
## ZupposOfLiePRingNC
##
ZupposOfLiePRingNC := function( L )
    local R, d, n, z, A, i, j, D, a, e, b, B, SL, GL;

    # set up and find autos
    R := FindAutos(L);
    d := MinimalGeneratorNumberOfLiePRing(L);
    n := DimensionOfLiePRing(L);
    z := p^0;
    R.auto := R.auto{[1..d]}{[1..d]} * z;

    # get matrix of indeterminates 
    A := MutableNullMat(n,n);
    for i in [1..n] do
        for j in [1..n] do
            A[i][j] := Indeterminate(Rationals, Concatenation("A",
                        String(i), String(j)));
        od;
    od;

    # run some checks 
    b := HasBlockForm(R.auto, A);
    D := Indeterminate(Rationals, "D");
    a := DeterminantMatDivFree(R.auto);
    GL := (R.eqns = [D*a-1] or R.eqns = [1-D*a]);
    SL := (R.eqns = [a-1] or R.eqns = [1-a]);

    # the block case
    if GL and IsList(b) then return GLZuppos(b); fi;

    # block case with one diagonal entry pm 1, group is C_2 x C_{p-1}
    if GL and 
      (R.auto = [ [ D*A[2][2], 0 ], [0, A[2][2]] ]*z or
       R.auto = [ [ D*A[2][2], A[1][2] ], [0, A[2][2]] ]*z or
       R.auto = [ [ A[1][1], 0 ], [0, A[1][1]*D] ]*z or
       R.auto = [ [ A[1][1], A[1][2] ], [0, A[1][1]*D] ]*z) then 
        return 3*Q2 + (1-Q2)* D1;
    fi;

    # A[1][1]^3 = 1, group is C_3 x C_{p-1} or C_{p-1} (see D9)
    if GL and R.auto = [ [ D^2*A[2][2]^2, 0 ], [0, A[2][2]] ]*z then 
        return 4*Q3*D9 + (1-Q3*D9)*D1;
    fi;

    # group is C_{p-1}
    if GL and R.auto = [ [ A[1][1], A[1][2] ], [0, A[1][1]^2] ]*z then 
        return GLZuppos([1]);
    fi;

    # group is C_{p-1}
    if GL and R.auto = [ [ A[1][1], 0 ], [A[2][1], 1] ]*z then 
        return GLZuppos([1]);
    fi;

    # block 2
    if GL and (R.auto = [ [ A[1][1], A[1][2], 0 ], 
                          [ A[2][1], A[2][2], 0 ], 
                          [ 0, 0, A[1][1]*A[2][2]-A[1][2]*A[2][1] ] ]*z 
           or  R.auto = [[ A[2][2]*A[3][3]-A[2][3]*A[3][2], 0, 0 ],
                         [0, A[2][2], A[2][3]],
                         [0, A[3][2], A[3][3]]]*z) then 
        return GLZuppos([2]);
    fi;

    # D^2 A[2][2]^2 A[3][3]^2 = 1, hence A[1][1] = pm 1, 
    # group is C_2 x C_{p-1}^2
    if GL and R.auto = [ [ D*A[2][2]*A[3][3], 0, A[1][3] ], 
                         [ 0, A[2][2], A[2][3] ], 
                         [ 0, 0, A[3][3] ] ] * z then 
        return 7*Q2 + (1-Q2)*D1*(q+1);
    fi;

    # blocks 1,1
    if GL and R.auto = [[A[1][1], A[1][2], A[1][3]],
                        [0, A[2][2], A[2][3]],
                        [0,0,A[1][1]^2]]*z then 
        return GLZuppos([1,1]);
    fi;
         
    # blocks 1,1, (blocks A[1][1], [[1,0],[A[3][2], A[3][3]])
    if GL and  R.auto = [ [ A[1][1], A[1][2], A[1][3] ], 
                          [ 0, 1, 0 ], 
                          [ 0, A[3][2], A[3][3] ] ]*z then 
        return GLZuppos([1,1]);
    fi;

    # block 1
    if GL and R.auto = [ [ 1, 0, 0 ], 
                         [ 0, A[2][2], A[2][3] ], 
                         [ 0, 0, A[2][2] ] ]*z then 
        return GLZuppos([1]);
    fi;

    # block 1
    if GL and R.auto = [ [ A[1][1], A[1][2], 0 ], 
                         [ 0, 1, 0],
                         [ 0, 0, A[1][1]^2 ] ]*z then 
        return GLZuppos([1]);
    fi;

    # A[2][2]^-1 = D^2 A[3][3]^3, hence block 1,1
    if GL and R.auto = [ [ D*A[3][3]^2, A[1][2], 0 ], 
                    [ 0, A[2][2], 0 ], 
                    [ 0, 0, A[3][3] ] ]*z then 
        return GLZuppos([1,1]);
    fi;

    # A[1][1] = pm 1, group is C_2 x C_{p-1}
    if GL and R.auto = [ [ D*A[2][2], 0, A[1][3] ], 
                         [ 0, A[2][2], A[2][3] ], 
                         [ 0, 0, 1 ] ]*z then 
        return 3*Q2 + (1-Q2)* D1;
    fi;

    # blocks 1,1
    if GL and R.auto = [ [ A[1][1], 0, A[1][3] ], 
                         [ 0, A[1][1]*A[3][3], A[2][3] ], 
                         [ 0, 0, A[3][3] ] ]*z then 
        return GLZuppos([1,1]);
    fi;

    # blocks 1,1
    if GL and R.auto = [ [ A[1][1], A[1][2], A[1][3] ], 
                         [ 0, A[2][2], 0 ], 
                         [ 0, 0, A[1][1]*A[2][2] ] ]*z then 
        return GLZuppos([1,1]);
    fi;

    # blocks 1,2
    if GL and  R.auto = [ [ A[1][1], A[1][2], 0, A[1][4] ], 
                          [ A[2][1], A[2][2], 0, A[2][4] ], 
                          [ 0, 0, A[1][1]*A[2][2]-A[1][2]*A[2][1], A[3][4] ],
                          [ 0, 0, 0, A[4][4] ] ]*z then 
        return GLZuppos([1,2]);
    fi;

    # block 2
    if GL and R.auto = [ [ A[3][3]*A[4][4]-A[3][4]*A[4][3], A[1][2], 
                                  A[3][2]*A[4][3]-A[3][3]*A[4][2], 
                                  A[3][2]*A[4][4]-A[3][4]*A[4][2]], 
                    [ 0, 1, 0, 0 ], 
                    [ 0, A[3][2], A[3][3], A[3][4]],
                    [ 0, A[4][2], A[4][3], A[4][4]]]*z then 
        return GLZuppos([2]);
    fi;

    # block 1
    if SL and R.auto = [ [ A[1][1], A[1][2] ], [ 0, A[2][2] ] ]*z then 
        return GLZuppos([1]);
    fi;

    # block 1
    if SL and R.auto = [ [ A[1][1], 0 ], [ 0, A[2][2] ] ]*z then 
        return GLZuppos([1]);
    fi;

    # A[1][1] = pm 1, group is C_2
    if SL and R.auto = [ [ A[1][1], A[1][2] ], [ 0, A[1][1] ] ]*z then 
        return Q2;
    fi;

    #  A[2][2]^3 = 1, group is C_3
    if SL and R.auto = [ [ A[2][2]^2, 0 ], [ 0, A[2][2] ] ]*z then 
        return Q3*D9;
    fi;

    # 5.23 (compare 5.46): group is C_{p^2-1}
    if GL and R.auto = [ [ 1, 0, 0 ], 
                         [ 0, A[2][2], A[3][2]*x ], 
                         [ 0, A[3][2], A[2][2]+A[3][2] ] ]*z then 
        return D1+D2;
    fi;

    # 5.46: see 5.23: group is C_{p+1}
    if SL and R.auto = [ [ A[1][1], A[2][1]*x ], 
                         [ A[2][1], A[1][1]+A[2][1] ] ]*z then 
        return Q2 + D2;
    fi;

    # block 1
    if SL and  R.auto = [ [ A[1][1], A[1][2], A[1][3]], 
                          [ 0, 1, A[1][2]*A[3][3] ],
                          [ 0, 0, A[3][3] ]]*z then 
        return GLZuppos([1]);
    fi;

    # group is trivial
    if R.eqns = [] and
       R.auto = [[1, A[1][2]], [0, 1]]*z then 
        return 0;
    fi;

    # group is C_{p-1}
    if R.eqns = [D*A[2][2]-1, A[1][2]*x] and
       R.auto = [[1, A[1][2]], [0, A[2][2]]]*z then 
        return GLZuppos([1]);
    fi;

    # group is C_2^2
    if R.eqns = [A[2][2]^2-1, D^2-1] and
       R.auto = [[D*A[2][2], D*A[2][4]], [0, A[2][2]]]*z then 
        return 3*Q2;
    fi;

    # group is C_2^2
    if R.eqns = [A[2][2]^2-1, D^2-1] and
       R.auto = [[D*A[2][2], D*A[2][4]*w], [0, A[2][2]]]*z then 
        return 3*Q2;
    fi;

    # group is C_2^2
    if R.eqns = [A[2][2]^2-1, D^2-1, A[1][2]*w-D*A[2][4]] and 
       R.auto = [[D*A[2][2], A[1][2]], [0, A[2][2]]]*z then 
           return 3*Q2;
    fi;

    # block 1
    if R.eqns = [A[1][1]^2*A[2][2]-1] and 
       R.auto = [[A[1][1], A[1][2]], 
                 [0,       A[2][2]]]*z then 
        return GLZuppos([1]);
    fi;

    # group is C_2 or C_4
    if R.eqns = [A[1][1]^4-1] and 
       R.auto = [[A[1][1], A[1][2]], 
                 [0,       A[1][1]^2]]*z then 
        return Q2;
    fi;

    # blocks 1,1
    if R.eqns = [D*A[3][3]-1, A[1][1]*A[2][2]-1] and 
       R.auto = [[A[1][1], A[1][2], A[1][3]],
                 [0,       A[2][2], A[2][3]],
                 [0,       0,       A[3][3]]]*z then 
         return GLZuppos([1,1]);
    fi;

    # block 1
    if R.eqns = [A[2][2]^2*A[3][3]-1] and 
       R.auto = [[A[2][2]*A[3][3], -A[2][2]*A[3][4], A[1][3]],
                 [0,                A[2][2],         A[2][3]], 
                 [0,                0,               A[3][3]]] * z then  
        return GLZuppos([1]);
    fi;

    # case 5.45: group has size 2*(p+1), with (p+1) cyclic
    #if R.eqns = [ A[1][2]^2*w-A[2][2]^2+1, D^2-1] and
    #   R.auto = [[D*A[2][2], A[1][2]], 
    #             [D*A[1][2]*w, A[2][2]]]*z then 
    if LibraryName(L) = "5.45" then 
        return 3*Q2 + (1-Q2)*D2*(1-D1);
    fi;

    # 5.22 
    # A23 = e w A32 and A33 = e A22 with e = +/-1 and w A32^2 + D = A33^2
    # yields p^2-1 options for (A32, A33) and these determine D
    # group has order 2*(p^2-1) with (p^2-1) cyclic;
    #if R.eqns = [ -A[3][2]^2*w^2+A[2][3]^2, 
    #               A[2][2]*A[3][2]*w-A[2][3]*A[3][3], 
    #               -A[3][2]*A[3][3]*w+A[2][2]*A[2][3], 
    #               A[2][2]^2-A[3][3]^2, 
    #               D*(A[3][2]^2*w-A[3][3]^2)+1, 
    #               D*A[3][3]*(A[2][2]*A[3][3]-A[2][3]*A[3][2])-A[2][2] ] and
    #   R.auto = [[ D*(A[2][2]*A[3][3]-A[2][3]*A[3][2]), 0, 0], 
    #             [ 0, A[2][2], A[2][3] ],
    #             [ 0, A[3][2], A[3][3] ]] * z then 

    if LibraryName(L) = "5.22" then 
       return 2*Q2 + (1-Q2)*(D1+D2); 
    fi;
 
    # 5.6 is a closely related to a symplectic group and 
    # has order p^4*(p-1)*(p^2-1)*(p^4-1)
    if LibraryName(L) = "5.6" then 
        return D4 + D2*(q+3+q4)/4 + D1*(1-Q2)*(q^2+6*q+9+2*q4)/8 + 3*Q2;
    fi;

    return "case not found";
end;

ZupposOfLiePRing := function(L)
    local ff;
    ff := ZupposOfLiePRingNC(L);
    if IsString(ff) then return ff; fi;
    return SimplifyMyPorcPoly(ff);
end;

#############################################################################
##
## The Lie rings 5.42 and 5.20 have to be considered as special cases,
## since we need to split their families. The two cases are very similar.
##
MyZuppos542 := function()
    local a, b, c;

    # 5.42
    # if x = 1 then this is SL(2,p)
    # if x <> 1 then
    #   A22 <> 0 yields A21 = A12 = 0 and A11 A22 = 1, thus (p-1)
    #   A22 = 0 yields x A12 A21 = -1 and A11 = 0 and x=-1, thus (p-1)
    # return "SL(2,p) if x=1, 2(p-1) if x =-1, (p-1) otherwise";

    # x = 1 (group is isomorphic to SL(2,p))
    a := D2 + D1;

    # x = -1 (group has the form C_{p-1} split C_2)
    b := 2*Q2 + D1;

    # x <> 1,-1 (group is cyclic of the form C_{p-1}), (p-3)/2 cases
    c := (p-3)/2*D1;

    return SimplifyMyPorcPoly(a + b + c);
end;

MyZuppos520 := function()
    local a, b, c;

    # 5.20
    # if x = 1, then this is GL(2,p)
    # if x <> 1 then
    #   A33 <> 0 yields A32=A23=0 and D A22 A33 = 1, thus (p-1)^2
    #   A33 = 0 yields D A23 A32 = 1 and A22 = 0 and x=-1, thus (p-1)^2
    # return "GL(2,p) if x=1, 2(p-1)^2 if x=-1, (p-1)^2 otherwise";

    # x = 1 (group is isomorphic to GL(2,p))
    a := GLZuppos([2]);

    # x = -1 (group has the form C_{p-1}^2 split C_2)
    b := 1/2*Q2 + D1*(q+3)/2;

    # x <> 1,-1 (group is cyclic of the form C_{p-1}^2), (p-3)/2 cases 
    c := (p-3)/2*D1*(q+1);

    return SimplifyMyPorcPoly(a + b + c);
end;

#############################################################################
##
## NumberNormalPSylow(n) determines P(p^nq) for q <> p and n <= 5.
##
InstallMethod( NumberPNQNormalPSylow, [IsPosInt], function( n )
    local OLDORDER, L, l, a, i, b, c;

    OLDORDER := ORDER;
    ORDER := MonomialGrlexOrdering();

    L := LiePRingsByLibrary(n);
    a := 0;
    for i in [1..Length(L)] do
        l := LibraryName(L[i]);
        if l = "5.20" then 
            a := a + MyZuppos520();
        elif l = "5.42" then 
            a := a + MyZuppos542();
        else
            b := NumberOfLiePRingsInFamily(L[i]);
            c := ZupposOfLiePRing(L[i]);
            a := a + b*c;
        fi;
    od;

    # correction term for q = 2 and q = 3 
    if n = 5 then 
        a := a - 4 * P2 * Q3 - 3 * P3 * Q2;
    fi;

    ORDER := OLDORDER;

    return SimplifyMyPorcPoly(a);
end );

#############################################################################
##
## NumberNormalPSylowP(p,n) determines P(p^nq) for given p and q <> p.
##
NumberNormalPSylowP := function( p, n )
    local pr, nr, rs, i, G, A, iso, B, r, j;
    pr := Set(Flat(List([1..n], x -> Factors(Size(GL(x,p))))));
    pr := Filtered(pr, IsPrimeInt);
    pr := Filtered(pr, x -> x <> p);
    nr := NumberSmallGroups(p^n);
    rs := NullMat(nr, Length(pr));
    for i in [1..nr] do
        Print("process group ",i," of ",nr, "\n");
        G := SmallGroup(p^n, i);
        A := AutomorphismGroup(G);
        iso := IsomorphismPermGroup(A);
        B := Image(iso);
        r := RationalClasses(B);
        r := List(r, x -> Order(Representative(x)));
        for j in [1..Length(pr)] do
            rs[i][j] := Length(Filtered(r, x -> x=pr[j]));
        od;
    od;
    return rec(primes := pr, numbers := Sum(rs));
end;


